/*
 *
 * $Id: MIME.java,v 1.2 1996/01/19 01:52:30 iansmith Exp $
 * $Author: iansmith $
 * $Log: MIME.java,v $
 * Revision 1.2  1996/01/19  01:52:30  iansmith
 * Some documentation improvements in the javadoc comments.
 *
 * Revision 1.1  1996/01/15  18:46:37  iansmith
 * Initial revision
 *
 *
 */

package GIT.GVU.IMAP;

/* need the java lanaguage */
import java.lang.*;
import java.util.*;

/**
 * Class to hold tables and functions associated with MIME functions of the toolkit.
 * Generally, these are just some miscellaneous functions that other parts of the
 * toolkit use.
 *
 * @version $Id: MIME.java,v 1.2 1996/01/19 01:52:30 iansmith Exp $
 * @author Ian Smith
 */

public class MIME {
  /**
   * Thie function will convert an array of input data to an array of output data,
   * one unit at a time. A unit is 4 bytes of input data and 3 of output data.
   * It returns the number of bytes actually converted. This is usually 3, but might
   * be 2 or 1 if we are at the end of the data. You should be careful not to call 
   * this function when you are past EOF.  Note: the input data size should always
   * be a multiple of 4.
   * 
   * @param byte[] input the array of input
   * @param int inIndex index of the input byte array to read from (note: this must be 0 mod 4 but isn't checked)
   * @param byte[] output the array of output
   * @param int outIndex  index in the output array to write to (note: this must be 0 mod 3 )
   * @exception IMAPException if the inputted base 64 encoded data has bad characters or too many consecutive = characters
   * @exception InternalErr is thrown if the BASE64 decoder gets confused about its own state
   * @return int the number of input bytes converted. This will be a number from 1 to 4; usually 4.  It will be 1, 2, or 3 only at the EOF. Note: Just because its 4 doesn't mean its not EOF if the boundary of the file falls evenly on a 4 byte boundary.
   */
  public static int decodeUnitBase64(byte[] input, int inIndex, byte[] output, int outIndex) throws IMAPException, InternalErr{
    String locator=new String(input,0,inIndex,4);
    int count=0,i,p;
    char c;
    byte pos;
    /* zero output data originally */
    for (i=0; i<4; ++i) output[i+outIndex]=0;

    /* loop over each character */
    for (i=0; i<4; ++i) {
      c=(char)input[inIndex+i];
      /* where is it in the alphabet ? */
      p=alphabet.indexOf(c,0);
      pos=(byte)p;
      /* if pos is not there ...  */
      if (p==-1) {
	/* its an equal sign, then ok */
	if (c=='=') {
	  continue;
	} else {
	  /* its crap, bail out */
	  throw new IMAPException("Bad character in BASE64 encoding");
	}
      }
      /* update counter */
      ++count;
      /* on each character position do something different */
      switch (i) {
      case 0:
	/* highest order six bits .. shift left two and just plain shove it in */
	pos <<= 2;
	output[outIndex] = pos;
	break;
      case 1:
	/* split the low 4 from pos out into the high 4 of the middle output byte */
	output[outIndex+1]=(byte)( (pos & 15) << 4);
	/* now do the highest two of pos into the lowest 2 of  high byte */
	output[outIndex]+=  pos >> 4;
	break;
      case 2:
	/* split the high four from pos into the low 4 of middle byte */
	output[outIndex+1]+= pos >>2;
	/* the low 2 of pos into the high two of lowest order byte */
	output[outIndex+2]=(byte)( (pos & 3) << 6);
	break;
      case 3:
	/* just put all six of these into the low order byte */
	output[outIndex+2]+=pos;
	break;
      default:
	throw new InternalErr("Bad index value in BASE64 decoder");
      }
    }
    /* if count ==0 then they gave us all = and this is wrong according to spec */
    if (count==0) {
      throw new IMAPException("Bad formation of BASE64 code: four consecutive = chars");
    }
    return count;
  }
  /**
   *  This variable holds (via its position) the value of every character in the base 64 
   * encoding.
   */
  static String alphabet="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
  /** 
   * Given a string, give us the BASE64 decoding of it as a string.
   * 
   * @param String s input string to decode 
   * @exception InternalErr is thrown if the unit decoder gives us a bogus return value
   * @return a String with the decoded value or NULL if decode failed
   */
  public static String decodeBase64(String input) throws InternalErr  {
    StringBuffer sb=new StringBuffer(16000); /* need a big one! */
    StringTokenizer st=new StringTokenizer(input);
    byte rawData[],outData[];
    String copy;
    int converted=0,tmp,outpoint=0;
    boolean eof=false;

    try {
      /* while there are more lines */
      while (st.hasMoreTokens()==true) {
	/* read next line and add to buffer */
	sb.append(st.nextToken());
      }
      /* after this process we have removed all the carriage returns and linefeeds and */
      /*have a buffer which is just data */
      /* to a string */
      copy=sb.toString();
      /* allocate space for input bytes*/
      rawData=new byte[copy.length()];
      /* allocate space for output bytes... we overestimate to start with */
      outData=new byte[copy.length()]; 
      /* conver to raw bytes */
      copy.getBytes(0,copy.length(),rawData,0);
      /* convert until we run out*/
      while (converted<copy.length() && eof!=true) {
	/* convert 4 bytes or some number a little less */
	tmp=decodeUnitBase64(rawData,converted,outData,outpoint);
	/* add tmp to the number we have converted */
	converted+=tmp;
	/* if tmp is now not 4 then we are at the end of the data, signalled */
	/* by the = in the data */
	if (tmp!=4) { 
	  eof=true;
	}
	/* we need to figure out how many bytes to move output forward */
	switch (tmp) {
	case 0:
	  throw new InternalErr("Bad value returned from decode unit!");
	case 1:
	  /* just one byte */
	  outpoint++;
	  break;
	case 2:
	  /* two chars is two bytes */
	  outpoint+=2;
	  break;
	case 3:
	case 4:
	  /* three or four chars is three bytes */
	  outpoint+=3;
	  break;
	default:
	  throw new InternalErr("Bad index value in decode() routine");
	}
      }
      /* now done everything just buid a string */
      return new String(outData,0, 0, outpoint);
    } catch (IMAPException e) {
      /* base 64 was badly coded, return null */
      return null;
    }
  }
  /**
   * This table holds all the MIME types that I think the HTTP server knows about and their
   * extensions.  Key is MIME Type/MIME Subtype and value is filename extension to use
   * trigger that type.
   */
  static Hashtable mimeExtensions;
  /**
   * This function gets run to initialize the HashTable of mimeExtensions.
   */
  static {
    /* init the table */
    mimeExtensions=new Hashtable(75,0.8F);
    /* I got all of these values from the file mime.types which is in this */
    /* directory ... this is the file that the HTTP server looks at to do its */
    /* typing, so I figured if I just reversed it I could pick what format */
    /* the server would think a file is */
    mimeExtensions.put("applications/octet-stream","bin");
    mimeExtensions.put("applications/pdf","pdf");
    mimeExtensions.put("applications/pdf","pdf");
    mimeExtensions.put("applications/postscript","ps");
    mimeExtensions.put("applications/rtf","rtf");
    mimeExtensions.put("applications/x-mif","mif");
    mimeExtensions.put("application/x-csh","csh");
    mimeExtensions.put("application/x-dvi","dvi");
    mimeExtensions.put("application/x-hdf","hdf"); /* what's this? */
    mimeExtensions.put("application/x-latex","latex");
    mimeExtensions.put("application/x-netcdf","cdf"); /*what's this */
    mimeExtensions.put("application/x-sh","sh"); 
    mimeExtensions.put("application/x-tcl","tcl");  /* ugh! */
    mimeExtensions.put("application/x-tex","tex"); 
    mimeExtensions.put("application/x-texinfo","texinfo"); 
    mimeExtensions.put("application/x-troff","tr"); 
    mimeExtensions.put("application/x-troff-man","man"); 
    mimeExtensions.put("application/x-troff-me","me"); 
    mimeExtensions.put("application/x-troff-ms","ms"); 
    mimeExtensions.put("application/x-wais-source","src"); 
    mimeExtensions.put("application/zip","zip"); 
    mimeExtensions.put("application/x-bcpio","bcpio"); 
    mimeExtensions.put("application/x-cpio","cpio"); 
    mimeExtensions.put("application/x-gtar","gtar"); 
    mimeExtensions.put("application/x-shar","shar"); 
    mimeExtensions.put("application/x-sv4cpio","sv4cpio"); 
    mimeExtensions.put("application/x-sv4crc","sv4crc"); 
    mimeExtensions.put("application/x-tar","tar"); 
    mimeExtensions.put("application/x-ustar","ustar"); 
    mimeExtensions.put("audio/basic","au"); 
    mimeExtensions.put("audio/x-wav","wav"); 
    mimeExtensions.put("audio/x-aiff","aiff"); 
    mimeExtensions.put("image/gif","gif");  /* this shouldn't occur */
    mimeExtensions.put("image/jpeg","jpg");  /* this shouldn't occur */
    mimeExtensions.put("image/tiff","tiff");
    /* DIdn't put this in because of sun raster format... */
    /* mimeExtensions.put("image/x-cmu-raster","ras"); */
    mimeExtensions.put("image/x-portable-anymap","pnm");
    mimeExtensions.put("image/x-portable-bitmap","pbm");
    mimeExtensions.put("image/x-portable-pixmap","ppm");
    mimeExtensions.put("image/x-portable-graymap","pgm");
    mimeExtensions.put("image/x-rgb","rgb");
    mimeExtensions.put("image/x-xbitmap","xbm");
    mimeExtensions.put("image/x-xpixmap","xpm");
    mimeExtensions.put("image/x-xwindowdump","xwd");
    mimeExtensions.put("text/html","html"); /*this shouldn't happen */
    mimeExtensions.put("text/plan","txt"); /*this shouldn't happen */
    mimeExtensions.put("text/richtext","rtx");
    mimeExtensions.put("text/tab-separated-values","tsv");
    mimeExtensions.put("text/x-setext","etx");
    mimeExtensions.put("video/quicktime","mov");
    mimeExtensions.put("video/mpeg","mpeg");
    mimeExtensions.put("video/x-msvideo","avi");
    mimeExtensions.put("video/x-sgi-movie","movie");
  }
  /**
   * Find a filename extension, given a mime type/subtype.
   * 
   * @param String mt MIME Type/Subtype
   * @return String with the extension (note: if we don't know the MIME type/subtype you give us we return "")
   */
  static public String findExtension(String mt) {
    /* make sure we don't get screwed by the case */
    String realKey=mt.toLowerCase();
    /* test for not there */
    if (mimeExtensions.containsKey(realKey)==false) {
      /* not there, don't bother */
      return "";
    }
    /* it is there, so just return its value */
    return (String)mimeExtensions.get(realKey);
  }
  /**
   * Utility function to convert a vector of body structure information into a 
   * BodyPart object. Note: This function might end up being called recursively
   * if the body structure ends up being multipart.  Note: The message number
   * Is the one associated with the FETCH command and may be deceptive w.r.t.
   * encapsulated RFC822 mail.
   *
   * @param Vector v vector of body part information
   * @param String p base part  string ... Note: on first call this is ""
   * @param int n part number of this part ... Note: on first call this is 1
   * @param Message m message object associated with this body part
   * @param int msgno number of the message we are constructing
   */
  static BodyPart structureToPart(Vector v,String p, int n, Message m, int msgno) 
                throws InternalErr  {
    int lines=-1;
    int size=-1;
    String type,subtype,tmp;
    BodyPart bp;
    Vector props,flags,envelope,bodystruct;
    Message encapsulated;

    /* ok, the big test is to test the element 0: if its a vector this is multipart */
    if (v.elementAt(0) instanceof Vector == true ) {
      /* it is a multipart, return the result of forming a multipart body part */
      return formMultipart(v,p+(new Integer(n)).toString(),m,msgno);
    } 
    /* get the type */
    type=(String)v.elementAt(0);
    /* get the subtype*/
    subtype=(String)v.elementAt(1);
    /* if its text, we need the number of lines */
    if (type.equalsIgnoreCase("TEXT")) {
      try {
	/* convert it? */
	lines=Integer.valueOf((String)v.elementAt(7),10).intValue();
      } catch (NumberFormatException e) {
	/* bogus ! */
	throw new InternalErr("Unable to find a number for the size of message in lines");
      }
    }
    /* try to get the size in bytes */
    try {
      /* convert it? */
      size=Integer.valueOf((String)v.elementAt(6) ,10).intValue();
    } catch (NumberFormatException e) {
      /* bogus ! */
      throw new InternalErr("Unable to find a number for the size of message in bytes");
    }
    /* check for possible null props */
    if (v.elementAt(2) instanceof String == true) {
      tmp=(String)v.elementAt(2);
      if (tmp.equals("NIL")==false) {
	throw new InternalErr("BOGUS! Found a string where an attribute list should be");
      }
      props=null;
    } else {
      props=(Vector)v.elementAt(2);
    }
    /*XXX We should be handling the MD5 field */
    /* its not multipart ... so just return the simple case */
    bp=new BodyPart(type,subtype,props,
                   (String)v.elementAt(3),(String)v.elementAt(4),
                   (String)v.elementAt(5),size,lines,
                   p+(new Integer(n)).toString(),m);

    /* check to see if its a RFC822 encapsulation */
    if (bp.getMIMEType().equalsIgnoreCase("MESSAGE")==true) {
      if (bp.getMIMESubType().equalsIgnoreCase("RFC822")==true) {
	/* build the inside message */
	flags=null;
	envelope=(Vector)v.elementAt(7);
	bodystruct=(Vector)v.elementAt(8);
	/*note: the message that is encapsulated is associated with 
	 * the mailbox of its parent.
	 */
	encapsulated=Mailbox.createMessage(flags,envelope,
                                           bodystruct,size,
                                           m.getMailbox(),
                                           bp.getPartNumber() + ".",msgno);
	/* set the embedded message */
	bp.setEmbeddedMessage(encapsulated);
      }
    }
    /* send it back to caller */
    return bp;
  }
  /** 
   * This function converts a multipart structure into a BodyPart.
   * 
   * @param Vector v a vector of structure information from the FETCH call
   * @param String base part number of this message. Note:  on first call, this is ""
   * @param Message m message object associated with this body part
   * @param int msgno message number of this multipart body part
   */
  static BodyPart formMultipart(Vector v, String p, Message m, int msgno) throws InternalErr {
    String subtype;
    boolean foundIt=false;
    BodyPart bp,current,drag=null,first=null;
    int i,subtypeElement=0,counter=1;
    
    /* loop over all the elements looking for first that isn't a vector */
    for (i=0;  i<v.size(); ++i) {
      if (foundIt==false) {
	if ( v.elementAt(i) instanceof Vector == false ) {
	  subtypeElement=i;
	  foundIt=true;
	}
      }
    }
    /* sanity check */
    if (subtypeElement==0) {
      throw new InternalErr("Unable to locate end of multipart parts in message");
    }
    /* we know how far to walk walk along, building a bp every time */
    /* and linking them all together .*/
    for (i=0; i<subtypeElement; ++i) {
      /* create  a new bp */
      current=structureToPart((Vector)v.elementAt(i),
                              p,counter,m,msgno);
      /* update counter */
      counter++;
      /* have we figured out the first one yet */
      if (first==null) {
	first=current;
      }
      /* now, we are dragging a pointer behind us, so we can link this guy to  */
      /* his neighbor */
      if (drag!=null) {
	drag.setSibling(current);
      }
      /* update the drag pointer */
      drag=current;
    }
    /* when we get out of the loop, we don't need to update the last guy */
    /* because we default his next sibling to null at creation time */

    /* ok extract the string that is the subtype */
    subtype=(String)v.elementAt(subtypeElement);

    /* construct the bp */
    bp=new BodyPart("MULTIPART",subtype,(Vector)v.elementAt(subtypeElement+1), 
                    "NIL","NIL","NIL",-1,-1,"",m);
    /* set the first child */
    bp.setFirstChild(first);

    /* we're done! */
    return bp;
  }
}
