/*
 * $Id: MailAddress.java,v 1.1 1996/01/08 05:56:54 iansmith Exp $
 * $Author: iansmith $
 * $Log: MailAddress.java,v $
 * Revision 1.1  1996/01/08 05:56:54  iansmith
 * Initial revision
 *
 *
 */

package GIT.GVU.IMAP;
import java.lang.*;
import java.util.*;

/**
 * Class to hold a single email address in RFC 1730 format. 
 * 
 * @version $Id: MailAddress.java,v 1.1 1996/01/08 05:56:54 iansmith Exp $
 * @author Ian Smith
 */

public class MailAddress {
  /**
   * Construct an MailAddress from the four strings present in the IMAP 4
   * protocol.
   * 
   * @param string pn personal name
   * @param string sr source route
   * @param string mn mailbox name
   * @param string hn host name
   */
  public MailAddress(String pn, String sr, String mn, String hn) {
    if (pn.equals("NIL")==true) {
      personalName=null;
    } else {
      personalName=pn;
    }
    if (sr.equals("NIL")==true) {
      sr=null;
    } else {
      sourceRoute=sr;
    }
    mailboxName=mn;
    hostname=hn;
  }
  /**
   * String to hold the personal name of a MailAddress
   */
  String personalName;
  /**
   * String to hold the source route of a MailAddress
   */
  String sourceRoute;
  /**
   * String to hold the mailbox name of a MailAddress
   */
  String mailboxName;
  /**
   * String to hold the hostname of a MailAddress. 
   */
  String hostname;
  /**
   * Fetch the personal name of a MailAddress. Note: This might be null
   * if no personal name was in the message.
   * 
   * @return string with the real name of the address i.e. "Ian Smith"
   */
  public String getPersonalName() { return personalName;};
  /**
   * Fetch the mailbox name of a MailAddress
   * 
   * @return string with the mailbox name of the address i.e. "iansmith"
   */
  public String getMailboxName() { return mailboxName;};
  /**
   * Fetch the source route of a MailAddress. Note: Might be null if 
   * it wasn't in the mail message.
   * 
   * @return string with the real name of the address 
   */
  public String getSourceRoute() { return sourceRoute;};
  /**
   * Fetch the hostname of a MailAddress
   * 
   * @return string with the hostname of the address i.e. "cc.gatech.edu"
   */
  public String getHostname() { return hostname;};
  /**
   * Convert to a string to make debugging a little easier. 
   */
  public String toString() {
    return (getPersonalName() + " <" + getMailboxName() + "@" + getHostname() + ">");
  }
}
