/*
 *  $Id: MailboxName.java,v 1.2 1996/01/08 06:00:30 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: MailboxName.java,v $
 *  Revision 1.2  1996/01/08 06:00:30  iansmith
 *  RCS Keyword cleanup.
 *
 *  Revision 1.1  1996/01/06 19:50:49  iansmith
 *  Initial revision
 *
 */
package GIT.GVU.IMAP;
import java.lang.*;

/**
 * Class to hold informationa about the *names* of mailboxes on a server.
 *
 * @version $Id: MailboxName.java,v 1.2 1996/01/08 06:00:30 iansmith Exp $
 * @author Ian Smith
 */
public class MailboxName {
  /**
   * Construct a new mailbox name. You must supply the name of the mailbox
   * and two booleans (selectable and interesting). Note: Some servers may
   * not implement the interesting function and this 
   *
   * @param n string which is the name of the mailbox
   * @param s boolean which should be true if this box is SELECTable
   * @param i boolean which should be true if this box has been marked as interesting by the server
   */
  MailboxName(String n,boolean s,boolean i) {
    interesting=i;
    selectable=s;
    name=n;
  }
  /** 
   * This variable holds the value of the selectable boolean.
   */
  boolean selectable;
  
  /** 
   * This variable holds the value of the interesting boolean.
   */
  boolean interesting;
  /**
   * This variable holds the name of the mailbox 
   */
  String name;
  /**
   * This function returns the value of the selectable boolean
   *
   * @return boolean whether or not this box selectable
   */
  public boolean getSelectable() { return selectable;};
  /**
   * This function returns the value of the interesting boolean
   * Remember, servers are not required to implement this, so this may
   * always be false.
   * 
   * @return boolean whether or not this box interesting
   */
  public boolean getInteresting() { return interesting;};
  /**
   * This function returns the name of the mailbox. You'll need to be sure
   * understand how the naming of boxes works with the LIST command before
   * trying to interpret this. See RFC1730.
   * 
   * @return string which is the name of the mailbox
   */
  public String getName() { return name;};
}
