/*
 * $Id: Display.java,v 1.2 1996/01/19 02:00:57 iansmith Exp $
 * $Author: iansmith $
 * $Log: Display.java,v $
 * Revision 1.2  1996/01/19  02:00:57  iansmith
 * Documentation improvements in the javadoc comments.
 *
 * Fixed a bug in the line counting code which was off by one and you
 * sometimes missed the top line of a message.
 *
 * Tried to add support for stopping messages in the middle of them being
 * printed to the terminal, but that doesn't appear to be working.
 *
 * Added a space after the from line part in the headers.
 *
 * Added some code so the headers are display "right" (as I think they
 * should be). If a message is \Seen, but still \Recent w.r.t. the
 * server, I don't mark it as new.
 *
 * Revision 1.1  1996/01/15  18:52:37  iansmith
 * Initial revision
 *
 */

package GIT.GVU.IMAP.Tests;
import GIT.GVU.IMAP.*;
import java.lang.*;
import java.util.*;
import java.io.*;
/**
 * Class which implements some printing functions for the little mail reader.
 * 
 * @version $Id: Display.java,v 1.2 1996/01/19 02:00:57 iansmith Exp $
 * @author Ian Smith
 */
public class Display {
 /** 
   * Function to demo how to extract some of the important information
   * from a mailbox.
   * 
   * @param Mailbox m mailbox of interest
   * @param boolean printFlags true if you want to see the flags defined for this box
   * @exception InternalErr is propagated
   */
  static void printMailboxInfo(Mailbox m,boolean printFlags) throws InternalErr{
    Vector flags,permanentFlags;
    int i;
    String tmp;

    /* check for read only mailbox */
    if (m.isReadOnly()==true) {
      System.out.println("MAILBOX IS READ-ONLY");
    } 
    /* any mail in box? */
    if (m.getExists()==0) {
      System.out.println("No mail in mailbox");
    } else {
      /* is it just 1 message or is it a bunch of them ? */
      if (m.getExists()==1) {
	/* is it new? */
	if (m.getRecent()==1) {
	  System.out.println("1 Messaage and it's new");
	} else {
	  /* no its not new */
	  System.out.println("1 Message");
	}
      } else {
	/* if more than one message, print that and print how */
	/* many new messages */
	System.out.println(m.getExists() + " Messages" +
                         "  " + m.getRecent() + " New");
      }
    }
    /* only display the list of flags and permanent flags if the caller */
    /* requested it... this information is probably only of interest */
    /* to programmers */
    if (printFlags==true) {
      /* grab the flags and print greeting */
      flags=m.getFlags();
      System.out.println("Flags on this mailbox"); 
      /* loop over the whole vector */
      for(i=0; i<flags.size(); ++ i) {
	/* convert the ith element into a string */
	tmp=(String)flags.elementAt(i);
	System.out.println(tmp);
      }
      /* grab the permanent (modifiable) flags and print greeting*/
      /* note: this list might be empty! */
      permanentFlags=m.getPermanentFlags();
      System.out.println("Permanent Flags on this mailbox"); 
      /* loop over the whole vector */
      for(i=0; i<permanentFlags.size(); ++ i) {
	/* convert the ith element into a string */
	tmp=(String)permanentFlags.elementAt(i);
	System.out.println(tmp);
      }
    }
  }
  /** 
   * Utility function to print a status line.
   * 
   * @param Vector v vector of flags
   */
  static void displayStatus(Vector v) {
    /* check to see if its new */
    if (Util.stringInVector(v,"\\Seen")==false) {
      /*not seen, so its new */
      System.out.print("Unseen");
    } else {
      /* seen it */
      System.out.print("Seen");
    }
    /* check to see if we have replied to it */
    if (Util.stringInVector(v,"\\Answered")==true) {
      System.out.println(" Answered");
    }
    /* check to see if its been flagged for deletion */
    if (Util.stringInVector(v,"\\Deleted")==true) {
      System.out.println(" Deleted");
    }
    /* output a CR */
    System.out.print("\n");
  }
  /** 
   * Utility function to print a list of mail addresses from a vector.
   * 
   * @param Vector v vector of mail addresses you want shown
   */
  static void displayAddresses(Vector v) {
    int i;
    String tmp;
    MailAddress m;

    /* loop over all addresses */
    for (i=0; i<v.size(); ++i) {
      /* extract addr */
      m=(MailAddress)v.elementAt(i);
      /* check for possible unknown in the real name */
      if (m.getPersonalName()==null) {
	System.out.print(m.getMailboxName()+"@"+m.getHostname());
      } else {
	/* convert to string */
	tmp=m.toString();
	/* now print it */
	System.out.print(tmp);
      }
      /* if we are not at end, print a comma */
      if (i!=v.size()-1) {
	System.out.print(", ");
      }
    }
    /* done print a CR */
    System.out.print("\n");
  }
  /**
   * This is the function that displays a mail message.
   * 
   * @param Connection c  connection to work from 
   * @param int n message number
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   */
  public static void displayMessage(Connection c, int  n) throws InternalErr, IMAPException , java.io.IOException {
    Mailbox m=c.getMailbox();
    Message message;
    MailAddress addr;
    String tmp,tmp2,fname="UNKNOWN";
    Vector v,flags;
    boolean commonCase=false;
    int lineCount=0;
    FileOutputStream fs;

    /* These two lines cause the message cache to be updated (if necessary)
     * and the mail message to be put into message. 
     */
    m.FETCH((new Integer(n)).toString());
    message=m.getMessage(n);
    
    /* Display the message */
    System.out.println("\nMessage    : " + n);
    ++lineCount;
    /* 
     * The common case is that the From, Sender, and Reply-To lines
     * are all the same. If they are, then we are going to compress
     * them into just a From for clarity. Further, they are usually
     * just one element long in each one.
     */
    /* test for common case */
    if ((message.getFrom().size()==1) &&
        (message.getSender().size()==1) &&
        (message.getReplyTo().size()==1)) {
      /* all have one element, are they the same? */
      v=message.getFrom();
      tmp=((MailAddress)v.elementAt(0)).toString();
      v=message.getSender();
      tmp2=((MailAddress)v.elementAt(0)).toString();
      /* ok, tmp and tmp2 are the first one of from and sender */
      if (tmp.equals(tmp2)==true) {
	/* they are same, check reply to */
	v=message.getReplyTo();
	tmp2=((MailAddress)v.elementAt(0)).toString();
	if (tmp.equals(tmp2)==true) {
	  /* set the flag to tell us to do the simple thing */
	  commonCase=true;
	}
      }
    }
    /* now print the right choice... based on the commondCase */
    if (commonCase) {
      /* ok, they are all the same, just print the from line */
      System.out.print("From       : ");
      lineCount++;
      displayAddresses(message.getFrom());
    } else {
      /* this is the uncommon case.. print everything */
      /* show the from line */
      System.out.print("From       : ");
      displayAddresses(message.getFrom());
      lineCount++;
      /* show the sender line */
      System.out.print("Sender     : ");
      displayAddresses(message.getSender());
      lineCount++;
      /* show the reply-to line */
      System.out.print("Reply-To   : ");
      displayAddresses(message.getReplyTo());
      lineCount++;
    }
    /* print the To line */
    System.out.print("To         : ");
    displayAddresses(message.getTo());
    lineCount++;
    /* check for CC */
    if (message.getCC()!=null) {
      /* print the CC line */
      System.out.print("CC         : ");
      displayAddresses(message.getCC());
      lineCount++;
    }
    /* check for BCC */
    if (message.getBCC()!=null) {
      /* print the CC line */
      System.out.print("BCC        : ");
      displayAddresses(message.getBCC());
      lineCount++;
    }
    /* show the subject line */
    System.out.println("Subject    : " + message.getSubject());
    lineCount++;
    /* show the Date */
    System.out.println("Date       : " + message.getDate());
    lineCount++;
    /* if there is an in-reply-to show it */
    if (message.getInReplyTo()!=null) {
      System.out.println("In Reply To: " +message.getInReplyTo());
      lineCount++;
    }
    /* show the message id */
    System.out.println("Message ID : " + message.getMessageID());
    lineCount++;
    /* show the status */
    System.out.print("Status     : ");
    lineCount++;
    displayStatus(message.getFlags());
    /* show the MIME information */
    System.out.print("MIME type  : ");
    lineCount+=displayStructure(message.getBodyPart(),0,12);
    /* Blank line for body then grab the data */
    System.out.println();
    lineCount++;
    /* if we are non-multipart message and thus have only one part...*/
    if (message.getBodyPart().getFirstChild()==null) {
      /* make sure the cache is up to date */
      message.getBodyPart().FETCH();
      /* print it out */
      displayBody(message.getBodyPart().getData(),lineCount);
    } else {
      try {
	/* get a filename from the user */
	System.out.print("Filename to write MIME message to (no .html) : ");
	fname=Util.getLine();
	/* create a file */
	fs=new FileOutputStream(fname + ".html");
	/* if writeMIME () fails, there was a problem writing the filename */
	if (Driver.writeMIME(message.getBodyPart(),fs,fname,0)==false) {
	  System.out.println("Error writing out MIME message as HTML!");
	} else {
	  /* it worked, close stream */
	  fs.close();
	}
      } catch (java.io.IOException e) {
	System.out.println("Error interacting file " + fname + ".html for writing");
      }
    }
  }
  /**
   * Print message structure out on the terminal 
   * 
   * @param BodyPart bp beginning of the list of body parts to walk over
   * @param int indent amount to indent per level of nesting
   * @param int lineup amount to indent all lines to offset the display from the left margin
   */
  public static int displayStructure(BodyPart bp,int indent,int lineup) {
    int i,count=0;
    BodyPart tmp;

    if (indent!=0) {
      /* lineup with the text */
      for (i=0; i<lineup; ++i) {
	System.out.print(" ");
      }
      /* print out the indention */
      for (i=0; i<indent; ++i) {
	System.out.print(" ");
      }
    }
    /* print out type/subtype */
    System.out.println(bp.getMIMEType() + "/" + bp.getMIMESubType()+ " ("
                       + bp.getPartNumber() +")");;
    ++count;
    /* is it a multipart ?*/
    if (bp.getFirstChild()!=null) {
      /* it is, so grab the head of the list */
      tmp=bp.getFirstChild(); 
      /* loop over list */
      while (tmp!=null) {
	/* print it out but indented more */
	count+=displayStructure(tmp,indent+5,lineup);
	/* go to next child */
	tmp=tmp.getNextSibling();
      }
    }
    /* is it an encapsulation? */
    if (bp.getEmbeddedMessage()!=null) {
      /* yes, so go to work on it by extracting the first body part from the inside msg*/
      count+=displayStructure(bp.getEmbeddedMessage().getBodyPart(),  
                              indent+5,lineup);
    }
    return count;
  }
  /**
   * Print the body of a text message on the terminal
   *
   * @param String m message's text
   * @param int s how many header lines have been printed 
   * @exception java.io.IOException is propagated
   *
   */
  static public void displayBody(String m,int s) throws java.io.IOException {
    int linecount=s;
    StringTokenizer st=new StringTokenizer(m,"\n");
    String tmp,line;
    boolean quit=false;

    /* while we have tokens */
    while ((st.hasMoreTokens()) && (quit==false)) {
      /* grab a line and print it */
      tmp=st.nextToken();
      System.out.println(tmp);
      /* update how many lines we have printed */
      ++linecount;
      /* screen is full? */
      if (linecount>=23) {
	/* stop and wait for return */
	System.out.print("Press <RETURN> for more or <q> to Quit...");
	line=Util.getLine();
	/* check for a q to quit */
	if (line.equalsIgnoreCase("q")==true) {
	  quit=false;
	}
	/* reset line counter */
	linecount=0;
      }
    }
  }
  /**
   * Build a short form of a mail address that a mail message is from. 
   * 
   * @param Message m message of interest
   * @return a string with a shortened form of a name for this person (preferably the From)
   */
  public static String buildShortFrom(Message m) {
    MailAddress ma;
    Vector from, sender, replyTo;
    /* extract the three vectors */
    from=m.getFrom();
    sender=m.getSender();
    replyTo=m.getReplyTo();
    /* XXX We only look at the first component of these vectors... */
    /* its very unusual (although not impossible) to get multiple */
    /* senders of the same message XXX */

    /* first try to find a real name from one of the three */
    ma=(MailAddress)from.elementAt(0);
    if (ma.getPersonalName()!=null) {
      return ma.getPersonalName();
    }
    ma=(MailAddress)sender.elementAt(0);
    if (ma.getPersonalName()!=null) {
      return ma.getPersonalName();
    }
    ma=(MailAddress)replyTo.elementAt(0);
    if (ma.getPersonalName()!=null) {
      return ma.getPersonalName();
    }
    /* ok, no personal name, try to construct a name from the email addr of from */
    ma=(MailAddress)from.elementAt(0);
    return (ma.getMailboxName()+"@"+ma.getHostname());
  }
  /**
   * Display the headers of your mail
   * 
   * @param Mailbox m Mailbox to show the headers of
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   */
  public static void displayHeaders(Mailbox m) throws InternalErr, IMAPException{
    int i;
    Message message;
    String mnum,line,from,subj,date;
    Vector fromVector;

    /* messages are 1-exists not 0-exists-1 */
    for (i=1; i<=m.getExists(); ++i) {
      /* eliminate old line */
      line="";
      /* conver to string */
      mnum=(new Integer(i)).toString();
      if (m.FETCH(mnum)==false) {
	System.out.println("Unable to fetch message " + mnum);
	return;
      }
      /* get the message */
      message=m.getMessage(i);
      /* flags */
      line=buildFlags(message);
      /* number */
      mnum=Util.makeWidth(mnum,4);
      line=line+mnum;
      /* from */
      from=buildShortFrom(message);
      from=Util.makeWidth(from,18);
      line=line+from+" ";
      /* date */
      date=message.getDate();
      date=Util.makeWidth(date,22);
      line=line+date+" ";
      /* subject */
      subj=message.getSubject();
      subj=Util.makeWidth(subj,28);
      line=line+subj;
      /* dump it out */
      System.out.println(line);
    }
  }
  /**
   * Build a list of flags for this mail message 
   * 
   * @param Message m message to extract flags from
   * @return a string with the flags (as characters, always 5 characters long)
   */
  static public String buildFlags(Message m) {
    String result;
    Vector flags=m.getFlags();

    /* does it even have flags? */
    if (m.getFlags()==null) {
      return "     ";
    }
    /* check for new mail */
    if (Util.stringInVector(flags,"\\Recent")==true) {
      /* extra check ... just because I like it this way */
      /* I don't like things being called new if I've seen them before */
      if (Util.stringInVector(flags,"\\Seen")==true) {
	result=" "; 
      } else {
	result="N"; /*N for new */
      }
    } else {
      result=" "; 
    }
    /* ok, has flags test our flags ... try seen first*/
    if (Util.stringInVector(flags,"\\Seen")==true) {
      result+="R";/* R for read */
    } else { 
      /* possibly new, so check for this */
      if (Util.stringInVector(flags,"\\Recent")==false) {
	result+="U";  /* unread and not new */
      } else {
	result+=" ";
      }
    }
    /* how about deleted */
    if (Util.stringInVector(flags,"\\Deleted")==true) {
      result=result + "D"; /* deleted */
    } else {
      result=result + " ";
    }
    /* how about replied? */
    if (Util.stringInVector(flags,"\\Answered")==true) {
      result=result+"A"; /* answered ? */
    } else {
      result=result+" ";
    }
    /* don't have any more for now ... add a blank*/
    result=result+" ";
    return result;
  }
}
