/*
 *  $Id: Util.java,v 1.5 1996/01/19 01:58:26 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: Util.java,v $
 * Revision 1.5  1996/01/19  01:58:26  iansmith
 * Documentation improvements in the javadoc comments.
 *
 * Revision 1.4  1996/01/15  18:52:20  iansmith
 * Moved a bunch of code out of this file and into the MIME class. I
 * moved the functions that formed the MIME BodyPart data structures
 * (structureToPart() and formMultipart() ).
 *
 * Revision 1.3  1996/01/10  22:04:56  iansmith
 * Wholesale modifications! First, I fixed a bug in the parser which was causing
 * it miss some tokens because of an off by one error in one of my substrings.
 *
 * Added a case to the parser to deal with literal strings.
 *
 * Added a function to read a line of text from the terminal.
 *
 * Major mod was the addition of two function which can convert body part
 * structures into a correctly formated data structure of body parts.
 * This is the structureToPart and formMultipart functions. They are
 * mutually recursive.  These two functions have the tough job of
 * looking at the strucure vectors, figuring out the mime structure
 * (and parts) and building body parts correctly linked to each AND
 * with the right MIME part numbers so they can be retreived later with
 * FETCH() on the BodyPart.
 *
 * Revision 1.2  1996/01/08  06:05:55  iansmith
 * Added the parsing code which parses the results of FETCH commands
 * (although it could be used to parse lots of things). This routine
 * is a recursive decent parser based on a lisp parser. It uses the
 * routines parseString() and buildList() to primarily do its work.
 * These routines are mutually recursive and parse atoms and lists
 * respectively. The parser uses findNonSpace() to find the next
 * character it needs to work with. They build up their return values
 * with the function formVector(). This function builds vectors of
 * size two with the parsed result in the 0th spot, and the next
 * character to consider for parsing as the second.
 *
 * Added removeParsingData() to build a cleaned up version of a
 * parse structure. This is because of the fact that formVector()
 * keeps holding onto the next character to parse. We eleminate
 * this extra crap with this function and give you a nicer parse
 * structure.
 *
 * Added parseAddresses which turns a vector of vector of strings
 * (the inner of which represent a mail address in RFC 1730 format)
 * into a vector of MailAddress objects.
 *
 *  Revision 1.1  1996/01/06 19:51:00  iansmith
 *  Initial revision
 *
 */


package GIT.GVU.IMAP;

/* Need all the java language of course */
import java.lang.*;
import java.util.*;

/**
 * Class to hold some miscellaneous utility routines. 
 *
 * @version $Id: Util.java,v 1.5 1996/01/19 01:58:26 iansmith Exp $
 * @author Ian Smith
 */
public class Util {
  /** 
   * Checks to see if the string s2 is the prefix of s1 
   * 
   * @param s1 string to check for a prefix
   * @param s2 prefix string
   * @return true if the s2 string is the prefix of s1
   */
  public static boolean lookingAt(String s1,String s2) {
    return (s1.regionMatches(0,s2,0,s2.length()));
  }
  /**
   * Checks a vector to see if the last element in the vector has a string
   * as the second token in the vector's last element. This is generally
   * to determine if the round trip reply is OK. 
   *
   * @param v the vector of results returned from the server
   * @param s the string to check for 
   * @return true if the string s is the second token in the last element of v
   */
  public static boolean replyToken(Vector v, String s) {
    /* extract last elemement */
    String t=(String)v.elementAt(v.size()-1);
    StringTokenizer st;
    String tmp;

    /* create tokenizer */
    st=new StringTokenizer(t);
    /* skip first token  */
    st.nextToken();
    /* put next token in tmp */
    tmp=st.nextToken();
    /* compare */
    return (tmp.equals(s));
  }
  /**
   * This function takes in a list (q.v. RFC1730) of strings and returns 
   * a list of the strings in the list. 
   * 
   * It assumes you have already removed the trailing ')' but have not
   * removed the initial '(' (which this code does remove).
   *
   * @param a list of strings formatted as in RFC 1730
   * @return a vector with those strings parsed out
   */
  public static Vector parseStringList(String val) {
    String s;
    StringTokenizer st;
    Vector result=new Vector();

    /* check to see if the value is empty */
    if (val.equals("(")==true) {
      /* we know there is no values, so just return the empty vector */
      return result;
    }
    /* get rid of first '(' */
    s=val.substring(2);
    /* make a tokenizer */
    st=new StringTokenizer(s);
    /* loop over all tokens */
    while (st.hasMoreTokens()==true) {
      result.addElement(st.nextToken());
    }
    /* finished, send it back to caller */
    return result;
  }
  /** 
   * This function determines if a given string is in a list of strings
   * that are encoded in a vector. This is useful if you consider Vector as
   * a set, and then this function tests for membership in the set.
   * 
   * @param v vector of strings
   * @param s is this string in the vector?
   * @return true if string is in the list of strings, false otherwise
   */
  public static boolean stringInVector(Vector v,String s) {
    int i;
    String tmp;

    /* loop over all elements of the vector */
    for (i=0; i<v.size();++i) {
      /* get the current string out*/
      tmp=(String)v.elementAt(i);
      /* are they the same? */
      if (tmp.equals(s)==true) {
	/* yep */
	return true;
      }
    }
    /* didn't find it , bail out with bad news*/
    return false;
  }
  /**
   * formats a string to a particular width
   * @param s string to format
   * @param w integer width
   * @return new string of the specified width
   */
  public static String makeWidth(String s,int w) {
    String tmp;
    int diff;
    byte ascii[];
    int i;

    /* maybe we got lucky and its already ok */
    if (w==s.length()) {
      return s;
    }
    /* maybe its too long */
    if (s.length()>w) {
      return (s.substring(0,w));
    }
    /* its too short ... how much do we need?*/
    diff=w-s.length();
    /* make an array of bytes to put spaces in */
    ascii=new byte[diff];
    /*fill in the new array */
    for (i=0; i<diff; ++i) {
      ascii[i]=' ';
    }
    /* make a string from the ascii */
    tmp=new String(ascii,0);
    /* return the concat */
    return (s+tmp);
  }
  /**
   * This functions finds the first non-string character in a string 
   * starting at a given position and returns its index. 
   *
   * @param s string to find first non-space in 
   * @param n index in s to start looking at
   * @return int index of first non-space, -1 if all characters are spaces
   * 
   */
  static int findNonSpace(String s,int n){
    int i=n; /* start at nth character */
    char c;
    boolean foundIt=false; /* terminating condition */
    /* loop over whole string */

    while ((i<s.length()) && (foundIt==false)) {
      /* get current character */
      c=s.charAt(i);
      /* is it a space? */
      if (c==' ') {
	/* its a space */
	++i;
      } else {
	/* it isn't jump out */
	foundIt=true;
      }
    }
    /* how did we get out .. note: > in case of going too far in string?*/
    if (i>=s.length()) {
      /* got out due to no spaces */
      return -1;
    }
    /* got out due to finding something */
    return i;
  }
  /**
   * Utility function to form a vector of two elements: a token and an
   * integer.
   * 
   * @param o object to put in the first slot of the vector
   * @param int i integer to convert to an Integer and put in second slot
   * @parma String s debugging string
   * @return vector of these two objects
   */
  static Vector formVector(Object o, int i,String s) {
    Vector result=new Vector(2);
    result.addElement(o); /* put first one in */
    result.addElement(new Integer(i)); /* convert second and add it */
    if (debugParser) {
      System.out.println("formVector: "+ s);
    }
    return result; /*send it home */
  }
  /**
   * Given a string and a position, return the string (atom or quoted
   * string) or list (as a vector) that this string parses into. If 
   * there are no more tokens in the string, it returns null. 

   * Note: What is actually returned is a vector of two elements. The first 
   * is the token (string or vector), the second is an Integer which is the 
   * next character position to process. If the returned value's first slot
   * is null, the second slot might be -1. If so, there is no more input.
   *
   * @param s string to parse
   * @param n starting position to start parsing at
   * @exception InternalErr  is thrown if the we unable to successfully parse a token
   * @return a string or a vector which is the parsed result of the input
   */
  static public Vector parseString(String s, int pos) throws InternalErr{
    int start,next=-1,i,target=-1;
    boolean found=false;
    int n;
    String num;

    if (debugParser==true ) {
      if ((pos>=s.length()) || (pos+1>=s.length())) {
	System.out.println("\n\tbuildList("+s+","+pos+")" + " About to go out of bounds...");
      } else {
	System.out.println("\n\tparseString("+s+","+pos+")" + " Looking at " + s.substring(pos,pos+1));
      }
    }

    /* find the first non-space character */
    start=Util.findNonSpace(s,pos);
    /*
     * Case 1: Nothing left in the string.
     */
    if (start==-1) {
      return (formVector(null,pos,"BAILING OUT WITH EOS"));
    }
    /* 
     * Case 2: Looking at a right paren and at end of a list
     * so effectively nothing left...
     */
    if (s.charAt(start)==')') {
      /* tell the user we have more input by making the 2nd one be start+1 */
      return (formVector(null,start+1,"BAILING OUT WITH RIGHT PAREN"));
    }
    /*
     * Case 3: Its the start of a list ... recurse down to build up a 
     * the structure
     */
    if (s.charAt(start)=='(') {
      /* call build list to build up all the elements */
      return (buildList(s,start));
    }
    /*
     * Case 4: Its an atom, but a quoted string
     */
    if (s.charAt(start)=='\"') {
      /* compute where the end of it is */
      next=s.indexOf("\"",start+1);
      /* sanity check */
      if (next==-1) {
	throw new InternalErr("Unable to find closing quote for quoted string");
      }
      /* now build the string in the middle of the two quotes*/
      return (formVector(s.substring(start+1,next),next+1,"FORMED QUOTED STRING" + s.substring(start+1,next)));
    }
    /*
     * Case 5: Literal String
     */
    if (s.charAt(start)=='{') {
      next=s.indexOf("}",start+1);
      /* sanity check */
      if (next==-1) {
	throw new InternalErr("Unable to locate match to a left curly brace during a parse");
      }
      /* extract the number */
      num=s.substring(start+1,next);
      try {
	n=Integer.valueOf(num,10).intValue();
      } catch (NumberFormatException e) {
	throw new InternalErr("Unable to find a number for the size of a string literal");
      }
      /* ok, got the number ... now retrieve that much of the input */
      return(formVector(s.substring(next+1,next+n+1),next+n+1,"FORMED LITERAL STRING: " + s.substring(next+1,next+n+1)));
    }
    /*
     * Case 6: Its a simple atom 
     */
    /* where does this atom end?... loop over the string from start to...*/
    for (i=start+1; (i<s.length()) && (found==false); ++i) {
      /* if its a space or either paren, bail out */
      if ((s.charAt(i)==' ') || (s.charAt(i)==')') || (s.charAt(i)=='(')) {
	/* bail plus set the target */
	found=true;
	target=i;
      }
    }
    /* if we get here and found==false then we are out of string */
    if (found==false) {
      /* just return up to the end */
      return(formVector(s.substring(start),next+1,"FORMED SIMPLE STRING " + s.substring(start)));
    }
    /* build the substring up to that space and return result */
    return (formVector(s.substring(start,target),target,"FORMED SIMPLE STRING " + s.substring(start,target)));
    
   }
  /** 
   * This function walks over a string to build up a list
   *
   * @param string s the string to walk over
   * @param int p where to start walking
   * @exception InternalErr is propagated
   * @return a vector of two elements: the first a vector with the list in it, the second is the next position to check 
   */
  static public Vector buildList(String s, int p) throws InternalErr {
    int start,done=-1; /* must init done to avoid dataflow problems */
    Vector v,result;
    Integer i;
    boolean ok=true;

    if (debugParser==true) {
      if ((p+1>=s.length()) || (p>=s.length())) {
	System.out.println("\n\tbuildList("+s+","+p+")" + " About to go out of bounds...");
      } else {
	System.out.println("\n\tbuildList("+s+","+p+")" + " Looking at: " + s.substring(p,p+1));
      }
    }
    /* where to put the stuff we build */
    result=new Vector(10,10);

    /* where is the first non-space which is not our first ( */
    start=Util.findNonSpace(s,p+1);
    /* loop until we run out of tokens ... and we'll jump out then*/
    while (ok==true) {
      /* get next potential token */
      v=parseString(s,start);
      /* are we done? */
      if (v.elementAt(0)==null) {
	/* we're done here ... figure out where stopped and put it in done*/
	i=(Integer)v.elementAt(1);
	done=i.intValue();
	ok=false;
      } else {
	/* not done, grab out the next element */
	result.addElement(v);
	if (debugParser==true) {
	  System.out.println("Just added an element to the list... size is now: " + result.size());
	}
	/* update the start position */
	i=(Integer)v.elementAt(1);
	start=i.intValue();
      }
    }
    /* return the result */
    return formVector(result,done,"FORMED A LIST OF ELEMENTS");
  }
  /**
   * This variable controls whether or not to debug the lisp-style 
   * parser for strings/lists.
   */
  static final boolean debugParser=false;
  /**
   * This function strips out the extra elements a list that are created
   * by the parsing routines.
   * 
   * @param vector v of 2 elements returned from the parsing routines
   * @return new object which is the first element of the two that came in plus a recursive decent of all children
   * 
   */
  public static Object removeParsingData(Vector v) {
    Object o;
    int i;
    Vector newData,origData;
    
    /* get first element */
    o=v.elementAt(0);
    /* is it a vector ? */
    if (o instanceof Vector) {
      /* convert o to a vector */
      origData=(Vector)o;
      /* allocate a new vector */
      newData=new Vector(origData.size());
      /* loop over all contents */
      for (i=0; i<origData.size(); ++i) {
	newData.addElement(removeParsingData((Vector)origData.elementAt(i)));
      }
      return newData;
    }
    /* not a vector, simply return it*/
    return o;
  }
  /**
   * This function takes a Vector which it assumes is a vector of vectors.
   * The inner vectors should be made of strings which are mail addresses
   * in RFC1730 format.  It returns a vector of MailAddress objects.
   *
   * @param Vector v vector of mail address vectors
   * @exception InternalErr is thrown if you pass in a vector which is not the right size to be a mail address
   * @return a vector of mail address objects 
   */
  public static Vector parseAddresses(Vector v) throws InternalErr {
    int i;
    Vector result=new Vector(10);
    MailAddress ma;
    Vector tmp;

    /* loop over all the elements of v to get all the mail addresses */
    for (i=0; i<v.size(); ++i) {
      /* get the ith element */
      tmp=(Vector)v.elementAt(i);
      /* sanity check */
      if (tmp.size()!=4) {
	throw new InternalErr("Wrong size vector to convert to mail address");
      }
      /* convert it to a mail address */
      ma=new MailAddress((String)tmp.elementAt(0),(String)tmp.elementAt(1),
                         (String)tmp.elementAt(2),(String)tmp.elementAt(3));
      /* add mail addr to result */
      result.addElement(ma);
    }
    /* send result to caller */
    return result;
  }
  /**
   * Utility function to read a line of text from the terminal.
   *
   * @return the string with the line of text read
   */
  public static String getLine() throws java.io.IOException {
    /* initially we are not done, obviously */
    boolean done= false;
    /* where we hold the byte as it is read */
    int inputVal;
    /* convert the inputchar to a character */
    char theChar;
    /* where we hold it */
    StringBuffer sbuff=new StringBuffer();
    
    /* just for good measure */
    System.out.flush();

    /* read until we get a CR */
    while (done==false) {
      /* read it */
      inputVal=System.in.read();
      /* convert it */
      theChar=(char)inputVal;
      /* check for a carriage return */
      if (theChar=='\n') {
	done=true;
      } else {
	/* append it */
	sbuff.append(theChar);
      }
    }
    return sbuff.toString();
  }
}

